"""Utility functions for LSST JupyterLab notebook environment."""

import os
import urllib
from pathlib import Path
from typing import Any, Optional, Union

import bokeh.io

_NO_K8S = False

try:
    from kubernetes import client, config
    from kubernetes.client.rest import ApiException
    from kubernetes.config.config_exception import ConfigException
except ImportError:
    _NO_K8S = True


def format_bytes(n: int) -> str:
    """Format bytes as text

    >>> format_bytes(1)
    '1 B'
    >>> format_bytes(1234)
    '1.23 kB'
    >>> format_bytes(12345678)
    '12.35 MB'
    >>> format_bytes(1234567890)
    '1.23 GB'
    >>> format_bytes(1234567890000)
    '1.23 TB'
    >>> format_bytes(1234567890000000)
    '1.23 PB'

    (taken from dask.distributed, where it is not exported)
    """
    if n > 1e15:
        return "%0.2f PB" % (n / 1e15)
    if n > 1e12:
        return "%0.2f TB" % (n / 1e12)
    if n > 1e9:
        return "%0.2f GB" % (n / 1e9)
    if n > 1e6:
        return "%0.2f MB" % (n / 1e6)
    if n > 1e3:
        return "%0.2f kB" % (n / 1000)
    return "%d B" % n


def get_hostname() -> str:
    """Utility function to return hostname or, failing that, "localhost"."""
    return os.environ.get("HOSTNAME") or "localhost"


def show_with_bokeh_server(obj: Any) -> None:
    """Method to wrap bokeh with proxy URL"""

    def jupyter_proxy_url(port: Optional[int] = None) -> str:
        """
        Callable to configure Bokeh's show method when a proxy must be
        configured.

        If port is None we're asking about the URL
        for the origin header.

        https://docs.bokeh.org/en/latest/docs/user_guide/jupyter.html
        """
        base_url = os.environ["EXTERNAL_INSTANCE_URL"]
        host = urllib.parse.urlparse(base_url).netloc

        # If port is None we're asking for the URL origin
        # so return the public hostname.
        if port is None:
            return host

        service_url_path = os.environ["JUPYTERHUB_SERVICE_PREFIX"]
        proxy_url_path = "proxy/%d" % port

        user_url = urllib.parse.urljoin(base_url, service_url_path)
        full_url = urllib.parse.urljoin(user_url, proxy_url_path)
        return full_url

    # The type: ignore is needed because the alternate form of notebook_url
    # (https://docs.bokeh.org/en/latest/docs/reference/io.html#bokeh.io.show)
    # isn't in bokeh's type hints.
    bokeh.io.show(obj=obj, notebook_url=jupyter_proxy_url)  # type:ignore


def get_pod() -> Optional[client.V1Pod]:
    """Try to get pod record.  If you don't have K8s or you can't use it
    (e.g. nubladov3 does not grant any K8s access to the Lab pod) then
    return None."""

    if _NO_K8S:
        return None
    try:
        config.load_incluster_config()
    except ConfigException:
        # We have the K8S libraries, but we don't have in-cluster config.
        return None
    api = client.CoreV1Api()
    namespace = "default"
    try:
        with open(
            "/var/run/secrets/kubernetes.io/serviceaccount/namespace", "r"
        ) as f:
            namespace = f.readlines()[0]
    except FileNotFoundError:
        pass  # use 'default' as namespace
    try:
        pod = api.read_namespaced_pod(get_hostname(), namespace)
    except ApiException:
        # Well, that didn't work.
        return None
    return pod


def get_node() -> str:
    """Extract node name from pod, or the empty string if we can't
    determine it."""
    pod = get_pod()
    if pod is not None:
        return pod.spec.node_name
    # In Nublado v3, we push this into the environment as K8S_NODE_NAME
    return os.environ.get("K8S_NODE_NAME", "")


def get_digest() -> str:
    """Extract image digest from pod, if we can.  Return the empty string
    if we cannot."""
    pod = get_pod()
    # Image ID looks like host/[project/]owner/repo@sha256:hash
    if pod is None:
        # In Nublado v3 we push the digest directly into the image spec
        image_id = os.environ.get("JUPYTER_IMAGE_SPEC", "")
    else:
        try:
            image_id = pod.status.container_statuses[0].image_id
        except Exception:
            image_id = ""
    try:
        return (image_id.split("@")[-1]).split(":")[-1]
    except Exception:
        return ""  # We will just return the empty string


def get_access_token(
    tokenfile: Optional[Union[str, Path]] = None, log: Optional[Any] = None
) -> str:
    """Determine the access token from the mounted location (nublado
    3/2) or environment (any).  Prefer the mounted version since it
    can be updated, while the environment variable stays at whatever
    it was when the process was started.  Return the empty string if
    the token cannot be determined.
    """
    tok = ""
    if tokenfile:
        # If a path was specified, trust it.
        tok = Path(tokenfile).read_text()
    else:
        jldir = "/opt/lsst/software/jupyterlab"
        # Try the default token paths: nublado3, then nublado2, then fall
        # back to the environment.
        tokenfiles = [
            f"{jldir}/secrets/token",
            f"{jldir}/environment/ACCESS_TOKEN",
        ]
        for tf in tokenfiles:
            token_path = Path(tf)
            try:
                tok = token_path.read_text()
                break
            except FileNotFoundError:
                pass
        if not tok:
            tok = os.environ.get("ACCESS_TOKEN", "")
    return tok



import os
import warnings
from typing import Optional

import pyvo
import pyvo.auth.authsession
import requests
from deprecated import deprecated

from .utils import get_access_token


def _get_tap_url() -> str:
    tapurl = os.getenv("EXTERNAL_TAP_URL")
    if not tapurl:
        tapurl = (os.getenv("EXTERNAL_INSTANCE_URL") or "") + (
            os.getenv("TAP_ROUTE") or "/api/tap"
        )
    return tapurl


def _get_obstap_url() -> str:
    obstapurl = os.getenv("EXTERNAL_OBSTAP_URL")
    if not obstapurl:
        obstapurl = (os.getenv("EXTERNAL_INSTANCE_URL") or "") + (
            os.getenv("OBSTAP_ROUTE") or "/api/obstap"
        )
    return obstapurl


def _get_datalink_url() -> str:
    return os.getenv("EXTERNAL_INSTANCE_URL", "") + "/api/datalink"


def _get_cutout_url() -> str:
    return os.getenv("EXTERNAL_INSTANCE_URL", "") + "/api/cutout"


def _get_auth() -> Optional[pyvo.auth.authsession.AuthSession]:
    tap_url = _get_tap_url()
    obstap_url = _get_obstap_url()
    s = requests.Session()
    tok = get_access_token()
    if not tok:
        return None
    s.headers["Authorization"] = "Bearer " + tok
    auth = pyvo.auth.authsession.AuthSession()
    auth.credentials.set("lsst-token", s)
    auth.add_security_method_for_url(_get_datalink_url(), "lsst-token")
    auth.add_security_method_for_url(_get_cutout_url(), "lsst-token")
    auth.add_security_method_for_url(tap_url, "lsst-token")
    auth.add_security_method_for_url(tap_url + "/sync", "lsst-token")
    auth.add_security_method_for_url(tap_url + "/async", "lsst-token")
    auth.add_security_method_for_url(tap_url + "/tables", "lsst-token")
    auth.add_security_method_for_url(obstap_url, "lsst-token")
    auth.add_security_method_for_url(obstap_url + "/sync", "lsst-token")
    auth.add_security_method_for_url(obstap_url + "/async", "lsst-token")
    auth.add_security_method_for_url(obstap_url + "/tables", "lsst-token")
    return auth


def get_tap_service() -> pyvo.dal.TAPService:
    #
    # This is not ideal, but warning appears because require pyvo does
    # not register uws:Sync and uws:Async.  It's harmless.  The broadness
    # of the warning is unfortunately necessary since pyvo just uses
    # warnings.warn():
    # https://github.com/astropy/pyvo/blob/
    # 81a50d7fd24428f17104a075bc0e1ac661ed6ea0/pyvo/utils/xml/elements.py#L418
    #
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        ts = pyvo.dal.TAPService(_get_tap_url(), _get_auth())
    return ts


def get_obstap_service() -> pyvo.dal.TAPService:
    #
    # This is not ideal, but warning appears because require pyvo does
    # not register uws:Sync and uws:Async.  It's harmless.  The broadness
    # of the warning is unfortunately necessary since pyvo just uses
    # warnings.warn():
    # https://github.com/astropy/pyvo/blob/
    # 81a50d7fd24428f17104a075bc0e1ac661ed6ea0/pyvo/utils/xml/elements.py#L418
    #
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        ts = pyvo.dal.TAPService(_get_obstap_url(), _get_auth())
    return ts


@deprecated(reason="Please use get_tap_service()")
def get_catalog() -> pyvo.dal.TAPService:
    return get_tap_service()


def retrieve_query(query_url: str) -> pyvo.dal.AsyncTAPJob:
    #
    # This is not ideal, but warning appears because require pyvo does
    # not register uws:Sync and uws:Async.  It's harmless.  The broadness
    # of the warning is unfortunately necessary since pyvo just uses
    # warnings.warn():
    # https://github.com/astropy/pyvo/blob/
    # 81a50d7fd24428f17104a075bc0e1ac661ed6ea0/pyvo/utils/xml/elements.py#L418
    #
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        atj = pyvo.dal.AsyncTAPJob(query_url, _get_auth())
    return atj



"""
Collection of utilities, formerly in rsp_jupyter_utils.lab and
rsp_jupyter_utils.helper
"""
from importlib.metadata import PackageNotFoundError, version

from .catalog import get_catalog, get_tap_service, retrieve_query
from .forwarder import Forwarder
from .log import IPythonHandler, forward_lsst_log
from .utils import (
    format_bytes,
    get_access_token,
    get_digest,
    get_hostname,
    get_node,
    get_pod,
    show_with_bokeh_server,
)

__version__: str
"""The application version string of (PEP 440 / SemVer compatible)."""

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    # package is not installed
    __version__ = "0.0.0"


__all__ = [
    "__version__",
    "Forwarder",
    "IPythonHandler",
    "format_bytes",
    "forward_lsst_log",
    "get_access_token",
    "get_catalog",
    "get_digest",
    "get_node",
    "get_pod",
    "get_tap_service",
    "retrieve_query",
    "get_hostname",
    "show_with_bokeh_server",
]



"""Tests for utility functions."""

from __future__ import annotations

from lsst.rsp import format_bytes


def test_format_bytes() -> None:
    assert format_bytes(1) == "1 B"
    assert format_bytes(1234) == "1.23 kB"
    assert format_bytes(12345678) == "12.35 MB"
    assert format_bytes(1234567890) == "1.23 GB"
    assert format_bytes(1234567890000) == "1.23 TB"
    assert format_bytes(1234567890000000) == "1.23 PB"
